/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.concurrent.ThreadPoolExecutor;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.message.ExchangeImpl;

/**
 * The JBI Message Exchange Acceptor. Dispatch the message to a JBI worker which
 * will use the {@link AbstractJBIListener} component implementation.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class MessageExchangeAcceptor extends Thread {

    /**
     * The component's logger
     */
    private Logger logger;

    private AbstractComponent component;

    private DeliveryChannel channel;

    private boolean running = false;

    private ThreadPoolExecutor workersPool;

    private GenericObjectPool jbiWorkersPool;

    /**
     * 
     * @param component
     * @param runtimeConfiguration
     * @param jbiPool
     */
    public MessageExchangeAcceptor(AbstractComponent component, GenericObjectPool jbiPool) {
        this.component = component;
        this.jbiWorkersPool = jbiPool;
        this.channel = component.getChannel();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        this.logger.log(Level.FINE, "Starting MessageExchangeAcceptor Thread " + this.getName());
        this.running = true;

        while (this.running) {

            try {
                // receiving a message
                ExchangeImpl exchange = new ExchangeImpl(this.acceptMessageExchange());
                this.logger.fine("Accepting a JBI message (ID : " + exchange.getExchangeId() + ")");
                this.workersPool.execute(createProcessor(exchange));
            } catch (MessagingException e) {
                // search for an interrupted exception
                Throwable cause = e;
                while (cause != null) {
                    cause = cause.getCause();
                    if (cause instanceof InterruptedException) {
                        this.running = false;
                        this.logger.log(Level.FINE, "Thread '" + this.getName()
                                + "' interrupted, stop accepting JBI messages.");
                    }
                }
                if (this.running) {
                    this.logger.log(Level.INFO, "This thread should be stopped now !");
                }
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    /**
     * 
     * @param logger
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * 
     * @return
     * @throws MessagingException
     */
    protected MessageExchange acceptMessageExchange() throws MessagingException {
        return this.channel.accept();
    }

    /**
     * 
     * @return
     */
    public ThreadPoolExecutor getWorkersPool() {
        return workersPool;
    }

    /**
     * 
     * @param workersPool
     */
    public void setWorkersPool(ThreadPoolExecutor workersPool) {
        this.workersPool = workersPool;
    }

    /**
     * Create the {@link MessageExchangeProcessor} that will handle the incoming
     * message exchange.
     * 
     * @return
     */
    protected MessageExchangeProcessor createProcessor(ExchangeImpl exchange)
            throws PEtALSCDKException {
        MessageExchangeProcessor processor = new MessageExchangeProcessor(this.component,
                this.jbiWorkersPool, exchange, logger);
        processor.setName(this.component.getContext().getComponentName() + " - JBI worker thread");
        return processor;
    }

	protected Logger getLogger() {
		return logger;
	}

	protected AbstractComponent getComponent() {
		return component;
	}

	protected GenericObjectPool getJbiWorkersPool() {
		return jbiWorkersPool;
	}
}
