/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005-2006 EBM WebSourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: SourceHelper.java 94 2006-03-26 17:11:05Z alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.net.URI;
import java.net.URLEncoder;
import java.util.Properties;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.api.exception.DocumentException;
import org.ow2.petals.component.framework.api.exception.FaultException;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Sample class used to manage the mapping between
 * <code>javax.xml.transform.Source</code> from String. the mapping is a quicly
 * "string" -> "<content>string</content>" conversion.
 * 
 * @author alouis - EBM WebSourcing
 * @since 1.0
 * 
 */
public final class SourceHelper {

    private static final String SOAPENV_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/";

    private static final String START_TAG = "<content>";

    private static final String END_TAG = "</content>";

    private static final int SOAP_FAULT_BUFFER_SIZE = 2048;

    private static final String UTF8_CHARFORMAT = "UTF-8";

    private final static ThreadLocal<DocumentBuilder> documentBuilderThreadLocal = new ThreadLocal<DocumentBuilder>() {

        @Override
        protected DocumentBuilder initialValue() {
            try {
                final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
                        .newInstance();
                documentBuilderFactory.setNamespaceAware(true);
                final DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
                return documentBuilder;
            } catch (ParserConfigurationException e) {
                throw new RuntimeException("Failed to create DocumentBuilder", e);
            }
        }
    };

    private final static ThreadLocal<Transformer> transformerWithoutXmlDeclarationThreadLocal = new ThreadLocal<Transformer>() {

        @Override
        protected Transformer initialValue() {
            try {

                final Transformer transformer = TransformerFactory.newInstance().newTransformer();
                Properties props = new Properties();
                props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
                transformer.setOutputProperties(props);
                return transformer;

            } catch (TransformerConfigurationException e) {
                throw new RuntimeException("Failed to create Transformer", e);
            }
        }
    };

    /**
     * Creates a new instance of {@link SourceHelper}
     * 
     */
    private SourceHelper() {
        super();
    }

    /**
     * Create a Source from String adding start and end tags.
     * 
     * @param msg
     * @return
     * @throws PEtALSCDKException
     * @deprecated Don't use this method. Prefer ${@link #createSource(Document)
     *  }
     */
    @Deprecated
    public static Source createContentSource(final String msg) throws PEtALSCDKException {
        return createSource(START_TAG + msg + END_TAG);
    }

    /**
     * Creates an XML fault with the stacktrace, message and the specified code
     * 
     * @param e
     *      exception, used to provide faultstring and detail
     * @param code
     *      code of the fault
     * @return A string containing the xml code that corresponds to the Soap
     *  Fault
     * @deprecated use method createSoapFaultDocument(final Throwable e, final
     *  QName qname) instead
     */
    @Deprecated
    public static String createSoapFault(final Throwable e, final String code) {
        return SourceHelper.createSoapFault(e, new QName(code));
    }

    /**
     * <p>
     * Creates an SOAP fault according to provided exception and code:<br>
     * <code>&lt;s:Fault xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultcode&gt;{org.objectweb.petals}ERROR&lt;/faultcode&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultstring&gt;mesage&lt;/faultstring&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultactor&gt;component&lt;/faultactor&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;detail&gt;<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
     * &nbsp;&nbsp;&nbsp;&lt;/detail&gt;<br>
     * &lt;s:Fault&gt;</code>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link DocumentException}:
     * <ul>
     * <li>the fault code is set with the value returned by {@link
     * DocumentException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by {@link
     * DocumentException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by {@link
     * DocumentException#getFaultActor()},</li>
     * <li>the detail is set with the exception document returned by {@link
     * DocumentException#getDocument()()}.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link FaultException}:
     * <ul>
     * <li>the fault code is set with the value returned by {@link
     * FaultException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by {@link
     * FaultException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by {@link
     * FaultException#getFaultActor()},</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * otherwise:
     * <ul>
     * <li>the fault code is set with the value of <code>errorCode</code>,</li>
     * <li>the fault string is set with the value returned by {@link
     * Exception#getMessage()},</li>
     * <li>the fault actor is set to <code>null</code>,</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * @param e
     *      An exception used to provide faultstring and detail
     * @param errorCode
     *      The QName of the fault code. It is not used if the exception is an
     *      instance of {@link FaultException} or {@link DocumentException}.
     * @return A string containing the xml code that corresponds to the Soap
     *  Fault, using the encoding UTF-8.
     * @deprecated use the method ${link {@link
     *  #createSoapFaultDocument(Throwable, QName)} instead.
     */
    @Deprecated
    public static String createSoapFault(final Throwable e, final QName errorCode) {
        try {
            Document faultDocument = createSoapFaultDocument(e, errorCode);

            return XMLUtil.createStringFromDOMDocument(faultDocument);

        } catch (Exception e1) {
            StringBuffer s = new StringBuffer(SourceHelper.SOAP_FAULT_BUFFER_SIZE);
            s.append("<s:Fault xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">");
            s.append("<faultcode>s:Server</faultcode>");
            s.append("<faultstring>").append(e1.getMessage()).append("</faultstring>");
            s.append("</s:Fault>");
            return s.toString();
        }

    }

    /**
     * <p>
     * Creates an SOAP fault, as a DOM document, according to provided exception
     * and code:<br>
     * <code>&lt;s:Fault xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultcode&gt;{org.objectweb.petals}ERROR&lt;/faultcode&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultstring&gt;mesage&lt;/faultstring&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultactor&gt;component&lt;/faultactor&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;detail&gt;<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
     * &nbsp;&nbsp;&nbsp;&lt;/detail&gt;<br>
     * &lt;s:Fault&gt;</code>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link DocumentException}:
     * <ul>
     * <li>the fault code is set with the value returned by {@link
     * DocumentException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by {@link
     * DocumentException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by {@link
     * DocumentException#getFaultActor()},</li>
     * <li>the detail is set with the exception document returned by {@link
     * DocumentException#getDocument()()}.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link FaultException}:
     * <ul>
     * <li>the fault code is set with the value returned by {@link
     * FaultException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by {@link
     * FaultException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by {@link
     * FaultException#getFaultActor()},</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * otherwise:
     * <ul>
     * <li>the fault code is set with the value of <code>errorCode</code>,</li>
     * <li>the fault string is set with the value returned by {@link
     * Exception#getMessage()},</li>
     * <li>the fault actor is set to <code>null</code>,</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * @param e
     *      An exception used to provide faultstring and detail
     * @param errorCode
     *      The QName of the fault code. It is not used if the exception is an
     *      instance of {@link FaultException} or {@link DocumentException}.
     * @return A string containing the xml code that corresponds to the Soap
     *  Fault
     */
    public static Document createSoapFaultDocument(final Throwable e, final QName errorCode) {

        QName faultCode = errorCode;
        String faultString = e.getMessage();
        if (faultString == null) {
            faultString = "null message";
        }
        String details = "";
        URI faultActor = null;
        if (e instanceof DocumentException) {
            faultCode = ((DocumentException) e).getFaultCode();
            faultActor = ((DocumentException) e).getFaultActor();
            final Document document = ((DocumentException) e).getDocument();
            if ((document != null) && (document.getDocumentElement() != null)) {
                try {
                    details = XMLUtil.parseToString(document.getDocumentElement());
                } catch (TransformerException e1) {
                    details = "An error occured when retrieving the exception document ("
                            + e1.getMessage() + ").";
                }
            }
        } else if (e instanceof FaultException) {
            faultCode = ((FaultException) e).getFaultCode();
            faultActor = ((FaultException) e).getFaultActor();

            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);

            details = sw.toString();

        } else {
            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);

            details = sw.toString();
        }

        // warning, the building of the document must be compatible DOM3 AND
        // DOM2
        final Document faultDocument = SourceHelper.documentBuilderThreadLocal.get().newDocument();

        final Element faultElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:Fault");
        // the NameSpace is not fill in an attribute with old XALAN DOM2
        // transformation
        if (!isDOM3(faultDocument)) {
            final Attr attr = faultDocument.createAttributeNS("http://www.w3.org/2000/xmlns/",
                    "xmlns:soapenv");
            attr.appendChild(faultDocument.createTextNode(SOAPENV_NAMESPACE));
            faultElement.setAttributeNode(attr);
        }
        faultDocument.appendChild(faultElement);

        final Element faultCodeElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:faultcode");

        if (faultCode != null) {
            if (faultCode.getNamespaceURI() != null) {
                if (faultCode.getNamespaceURI().equals(SOAPENV_NAMESPACE)) {
                    faultCodeElement.appendChild(faultDocument.createTextNode("soapenv:"
                            + faultCode.getLocalPart()));
                } else {
                    String faultCodePrefix = faultCode.getPrefix();
                    if (StringHelper.isNullOrEmpty(faultCodePrefix)) {
                        faultCodePrefix = "fcns";
                    }
                    final Attr attr = faultDocument.createAttributeNS(
                            "http://www.w3.org/2000/xmlns/", "xmlns:" + faultCodePrefix);
                    attr.appendChild(faultDocument.createTextNode(faultCode.getNamespaceURI()));
                    faultCodeElement.setAttributeNode(attr);
                    faultCodeElement.appendChild(faultDocument.createTextNode(faultCodePrefix + ":"
                            + faultCode.getLocalPart()));
                }
            } else {
                faultCodeElement
                        .appendChild(faultDocument.createTextNode(faultCode.getLocalPart()));
            }
        }

        final Element faultStringElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:faultstring");
        faultStringElement.appendChild(faultDocument.createTextNode(faultString));

        final Element faultDetailElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:detail");
        faultDetailElement.appendChild(faultDocument.createCDATASection(details));

        final Element faultActorElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:faultactor");
        if (faultActor != null) {
            faultActorElement.appendChild(faultDocument.createTextNode(faultActor.toString()));
        }

        faultElement.appendChild(faultCodeElement);
        faultElement.appendChild(faultStringElement);
        faultElement.appendChild(faultActorElement);
        faultElement.appendChild(faultDetailElement);

        return faultDocument;
    }

    /**
     * Create a Source from the String. The string must be in UTF-8 format,
     * otherwise, the Source is not created.
     * 
     * @param msg
     *      an UTF-8 String
     * @return the Source built from the String, null if not an UTF-8
     * @throws PEtALSCDKException
     * @deprecated use the method ${@link #createSource(Document)} instead.
     */
    @Deprecated
    public static Source createSource(final String msg) throws PEtALSCDKException {
        StreamSource source = new StreamSource();
        try {
            byte[] msgByte = msg.getBytes(UTF8_CHARFORMAT);
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (IOException e) {
            throw new PEtALSCDKException(e);
        }
        return source;
    }

    /**
     * Create a Source from the String. The String must be a valid XML String.
     * If an encoding is specified in the XML String header it is used to create
     * the XML bytes stream, otherwise the default UTF-8 encoding is used.
     * 
     * @param msg
     *      a valid xml String
     * @return the Source built from the XML String
     * @throws UnsupportedEncodingException
     *      If the named charset is not supported
     * @throws PEtALSCFException
     *      if an error occurs during Source creation
     * @deprecated use the method ${@link #createSource(Document))
     */
    @Deprecated
    public static Source createSourceFromString(final String msg) throws PEtALSCDKException {
        String charsetEncoding = XMLUtil.extractXmlEncoding(msg);
        return createSourceFromString(msg, charsetEncoding);
    }

    /**
     * Create a Source from the String with the given charset encoding. The
     * String must be a valid XML String.
     * 
     * @param xmlMessage
     *      a valid xml String message
     * @param charsetEncoding
     *      a supported encoding
     * @return the Source built from the XML String with the specified encoding
     * @throws UnsupportedEncodingException
     *      If the named charset is not supported
     * @throws PEtALSCFException
     *      if an error occurs during Source creation
     */
    private static Source createSourceFromString(final String xmlMessage, String charsetEncoding)
            throws PEtALSCDKException {
        StreamSource source = new StreamSource();
        try {
            byte[] msgByte = xmlMessage.getBytes(charsetEncoding);
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (IOException e) {
            throw new PEtALSCDKException(e);
        }
        return source;
    }

    /**
     * Create a Source from the String with a specified encoding format. If
     * charFormat is null or empty, UTF-8 will be used
     * 
     * @param msg
     *      The string to create source from
     * @param charFormat
     *      the encodig format
     * @return the Source built from the String, null if not an UTF-8
     * @deprecated use the method ${@link #createSource(Document)}
     */
    @Deprecated
    public static Source createSource(final String msg, final String charFormat)
            throws PEtALSCDKException {
        StreamSource source = new StreamSource();
        try {
            byte[] msgByte = msg.getBytes(StringHelper.isNullOrEmpty(charFormat) ? UTF8_CHARFORMAT
                    : charFormat);
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (IOException e) {
            throw new PEtALSCDKException(e);
        }
        return source;
    }

    /**
     * Create a Source from String adding start and end tags.
     * 
     * @param msg
     * @return
     * @throws PEtALSCDKException
     * @deprecated use the method ${@link #createSource(Document)}
     */
    @Deprecated
    public static Source createContentSourceWithURLEncoding(final String msg,
            final String charFormat) throws PEtALSCDKException {
        Source source;
        try {
            String encodedMsg = URLEncoder.encode(msg,
                    StringHelper.isNullOrEmpty(charFormat) ? "UTF-8" : charFormat);
            source = createSource(START_TAG + encodedMsg + END_TAG, charFormat);
        } catch (UnsupportedEncodingException e) {
            throw new PEtALSCDKException(e);
        }

        return source;

    }

    /**
     * Create a String from a Source. The String will be in UTF-8.
     * 
     * @param s
     * @return
     * @throws PEtALSCDKException
     * @deprecated use the method ${@link #createDocument(Source)}
     */
    @Deprecated
    public static String createString(final Source s) throws PEtALSCDKException {

        return SourceHelper.createString(s, UTF8_CHARFORMAT);
    }

    /**
     * Create a String from a Source with a given char encoding. The String will
     * be in UTF-8 if charFormat is null or empty. The source would be still
     * available.
     * 
     * @param s
     *      the source
     * @param charFormat
     *      the char encoding. e.g: UTF-8
     * @return
     * @throws PEtALSCDKException
     * @deprecated use the method ${@link #createDocument(Source)}
     */
    @Deprecated
    public static String createString(final Source s, String charFormat) throws PEtALSCDKException {

        String result = null;
        try {
            if (s instanceof StreamSource && ((StreamSource) s).getInputStream() != null) {
                final StreamSource ss = (StreamSource) s;
                final InputStreamForker isf = new InputStreamForker(ss.getInputStream());

                final InputStream isOne = isf.getInputStreamOne();
                ss.setInputStream(isf.getInputStreamTwo());

                byte[] buf = new byte[isOne.available()];
                BufferedInputStream bis = new BufferedInputStream(isOne);
                bis.read(buf);
                result = new String(buf, StringHelper.isNullOrEmpty(charFormat) ? UTF8_CHARFORMAT
                        : charFormat);
            } else {
                final StringWriter buffer = new StringWriter();
                final Result sresult = new StreamResult(buffer);
                SourceHelper.transformerWithoutXmlDeclarationThreadLocal.get()
                        .transform(s, sresult);
                result = buffer.toString();
            }
        } catch (IOException e) {
            throw new PEtALSCDKException(e);
        } catch (TransformerConfigurationException e) {
            throw new PEtALSCDKException(e);
        } catch (TransformerException e) {
            throw new PEtALSCDKException(e);
        }
        return result;
    }

    /**
     * Create a Source from a DOM document.
     * 
     * @param document
     *      A DOM document
     * @return the Source built from the DOM document
     */
    public static Source createSource(final Document document) {
        // normalize the document to assure the resolution of each node
        // namespace
        document.normalizeDocument();
        return new DOMSource(document);
    }

    /**
     * Create a DOM Document from a Source.
     * 
     * @param s
     *      the ${link {@link Source}
     * @return the DOM document
     * @throws PEtALSCDKException
     */
    public static Document createDocument(final Source s) throws PEtALSCDKException {

        try {
            final Document document = documentBuilderThreadLocal.get().newDocument();
            final DOMResult domResult = new DOMResult(document);
            SourceHelper.transformerWithoutXmlDeclarationThreadLocal.get().transform(s, domResult);
            return document;
        } catch (TransformerException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /**
     * Create a String from a Source without the root element 'content'.
     * 
     * @param s
     * @return the String in UTF-8 representation without the start and end tags
     * @throws PEtALSCDKException
     * @deprecated Don't use this method. Prefer ${@link #createDocument(Source)
     *  }
     */
    @Deprecated
    public static String createStringContent(final Source s) throws PEtALSCDKException {
        String content = createString(s);
        int index = content.indexOf(START_TAG);
        String result = content.substring(index + START_TAG.length());
        return result.substring(0, result.indexOf(END_TAG));
    }

    /**
     * Convert a Source to a String within UTF-8 format.
     */
    public static String convertToString(Source source) throws PEtALSCDKException {
        return convertToString(source, UTF8_CHARFORMAT);
    }

    /**
     * Convert a Source to a String with a given char encoding. The String will
     * be in UTF-8 if charFormat is null or empty.
     * 
     * @param s
     *      the source
     * @param charFormat
     *      the char encoding. e.g: UTF-8
     * @return the result String
     * @throws PEtALSCDKException
     */
    public static String convertToString(Source source, final String charFormat)
            throws PEtALSCDKException {
        String result = null;
        try {
            if (source instanceof StreamSource && ((StreamSource) source).getInputStream() != null) {
                InputStream inputStream = ((StreamSource) source).getInputStream();
                byte[] buf = new byte[inputStream.available()];
                BufferedInputStream bis = new BufferedInputStream(inputStream);
                bis.read(buf);
                result = new String(buf, StringHelper.isNullOrEmpty(charFormat) ? UTF8_CHARFORMAT
                        : charFormat);
            } else {
                final StringWriter buffer = new StringWriter();
                final Result sresult = new StreamResult(buffer);
                SourceHelper.transformerWithoutXmlDeclarationThreadLocal.get().transform(source,
                        sresult);
                result = buffer.toString();
            }
        } catch (IOException e) {
            throw new PEtALSCDKException(e);
        } catch (TransformerConfigurationException e) {
            throw new PEtALSCDKException(e);
        } catch (TransformerException e) {
            throw new PEtALSCDKException(e);
        }
        return result;
    }

    public static boolean isDOM3(Document document) {

        if (document == null) {
            return false;
        }

        final Field fields[] = document.getClass().getDeclaredFields();
        boolean result = false;
        for (Field field : fields) {
            if (field.equals("lookupPrefix")) {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Convert the DOM2 document to a StreamSource.
     * 
     * @throws PEtALSCDKException
     */
    public static Source convertDOM2ToSource(final Document dom2Document) throws PEtALSCDKException {

        final ByteArrayOutputStream bufferOut = new ByteArrayOutputStream();
        final Result sResult = new StreamResult(bufferOut);
        try {
            transformerWithoutXmlDeclarationThreadLocal.get().transform(
                    new DOMSource(dom2Document), sResult);
        } catch (TransformerException e) {
            throw new PEtALSCDKException("Error while transform DOM2 document to StreamSource", e);
        }

        // TODO (CDE): Use CircularBuffer to increase performances.
        final ByteArrayInputStream bufferIn = new ByteArrayInputStream(bufferOut.toByteArray());

        return new StreamSource(bufferIn);

    }
}
